/*-
 * Copyright (c) 2018 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package admin

import (
	"errors"
	"strconv"
	"strings"
	"time"
)

// Pong represents the response to the "ping" command.
type Pong struct {
	Time       time.Time
	CLIVersion string
}

// Ping encapsulates the "ping" command. error is non-nil when the
// response was not OK. In that case, the error is an
// UnexpectedResponse, which wraps the response from Varnish.
func (adm *Admin) Ping() (Pong, error) {
	pong := Pong{}
	resp, err := adm.Command("ping")
	if err != nil {
		return pong, err
	}
	if resp.Code != OK {
		badResp := UnexpectedResponse{Response: resp}
		return pong, badResp
	}
	answer := strings.Split(resp.Msg, " ")
	epoch, err := strconv.Atoi(answer[1])
	if err != nil {
		return pong, err
	}
	pong.Time = time.Unix(int64(epoch), 0)
	pong.CLIVersion = answer[2]
	return pong, err
}

// A State represents the status of the Varnish child process.
type State uint8

const (
	// Stopped indicates that the child was stopped regularly.
	Stopped State = iota

	// Starting indicates that the management process has begun
	// but not completed starting the child process.
	Starting

	// Running indicates that the child process is running
	// regularly.
	Running

	// Stopping indicates that the management process is shutting
	// down the child process regularly.
	Stopping

	// Died indicates that the child process process stopped
	// unexpectedly, and will be restarted by the management
	// process.
	Died
)

func (state State) String() string {
	switch state {
	case Stopped:
		return "stopped"
	case Starting:
		return "starting"
	case Running:
		return "running"
	case Stopping:
		return "stopping"
	case Died:
		return "died unexpectedly (being restarted)"
	default:
		panic("Invalid state")
	}
}

// Status encapsulates the "status" command. error is non-nil when the
// response was not OK. In that case, the error is an
// UnexpectedResponse, which wraps the response from Varnish.
func (adm *Admin) Status() (State, error) {
	resp, err := adm.Command("status")
	if err != nil {
		return Died, err
	}
	if resp.Code != OK {
		badResp := UnexpectedResponse{Response: resp}
		return Died, badResp
	}
	if strings.Contains(resp.Msg, "running") {
		return Running, nil
	}
	if strings.Contains(resp.Msg, "stopped") {
		return Stopped, nil
	}
	if strings.Contains(resp.Msg, "starting") {
		return Starting, nil
	}
	if strings.Contains(resp.Msg, "stopping") {
		return Stopping, nil
	}
	if strings.Contains(resp.Msg, "died") {
		return Died, nil
	}
	return Died, errors.New("unknown status")
}

// Start encapsulates the "start" command. error is non-nil when the
// response was not OK. In that case, the error is an
// UnexpectedResponse, which wraps the response from Varnish.
func (adm *Admin) Start() error {
	resp, err := adm.Command("start")
	if err != nil {
		return err
	}
	if resp.Code != OK {
		badResp := UnexpectedResponse{Response: resp}
		return badResp
	}
	return nil
}

// Stop encapsulates the "stop" command. error is non-nil when the
// response was not OK. In that case, the error is an
// UnexpectedResponse, which wraps the response from Varnish.
func (adm *Admin) Stop() error {
	resp, err := adm.Command("stop")
	if err != nil {
		return err
	}
	if resp.Code != OK {
		badResp := UnexpectedResponse{Response: resp}
		return badResp
	}
	return nil
}

// VCLLoad encapsulates the "vcl.load" command. error is non-nil when
// the response was not OK. In that case, the error is an
// UnexpectedResponse, which wraps the response from Varnish.
//
// Varnish usually returns status 106 (Param) when the load fails due
// to a compile failure, and the error message is in the wrapped
// Response.Msg.
//
// config is the configuration name under which the VCL instance can
// be referenced afterward. The name may not be already in use. path
// is file path at which the VCL source is located.
func (adm *Admin) VCLLoad(config string, path string) error {
	resp, err := adm.Command("vcl.load", config, path)
	if err != nil {
		return err
	}
	if resp.Code != OK {
		badResp := UnexpectedResponse{Response: resp}
		return badResp
	}
	return nil
}

// VCLUse encapsulates the "vcl.use" command. error is non-nil when
// the response was not OK. In that case, the error is an
// UnexpectedResponse, which wraps the response from Varnish.
//
// config is the configuration name previousy set with VCLLoad(), or a
// previous invocation of "vcl.load". On success, that config becomes
// the active VCL instance.
func (adm *Admin) VCLUse(config string) error {
	resp, err := adm.Command("vcl.use", config)
	if err != nil {
		return err
	}
	if resp.Code != OK {
		badResp := UnexpectedResponse{Response: resp}
		return badResp
	}
	return nil
}

// VCLDiscard encapsulates the "vcl.discard" command. error is non-nil
// when the response was not OK. In that case, the error is an
// UnexpectedResponse, which wraps the response from Varnish.
//
// config is the configuration name previousy set with VCLLoad(), or a
// previous invocation of "vcl.load". The config may not be the
// current active config. On success, that config is no longer
// available, and cannot become active via VCLUse() or "vcl.use".
func (adm *Admin) VCLDiscard(config string) error {
	resp, err := adm.Command("vcl.discard", config)
	if err != nil {
		return err
	}
	if resp.Code != OK {
		badResp := UnexpectedResponse{Response: resp}
		return badResp
	}
	return nil
}
