/*-
 * Copyright (c) 2018 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package log

import (
	"bytes"
	"strconv"
	"testing"
)

type atoUint32Exp struct {
	val uint32
	ok  bool
}

func TestAtoUint32(t *testing.T) {
	expMap := map[string]atoUint32Exp{
		"":           atoUint32Exp{val: 0, ok: false},
		"0":          atoUint32Exp{val: 0, ok: true},
		"1":          atoUint32Exp{val: 1, ok: true},
		"9":          atoUint32Exp{val: 9, ok: true},
		"10":         atoUint32Exp{val: 10, ok: true},
		"101":        atoUint32Exp{val: 101, ok: true},
		"1009":       atoUint32Exp{val: 1009, ok: true},
		"foo":        atoUint32Exp{val: 0, ok: false},
		"4294967290": atoUint32Exp{val: 4294967290, ok: true},
		"4294967291": atoUint32Exp{val: 4294967291, ok: true},
		"4294967292": atoUint32Exp{val: 4294967292, ok: true},
		"4294967293": atoUint32Exp{val: 4294967293, ok: true},
		"4294967294": atoUint32Exp{val: 4294967294, ok: true},
		"4294967295": atoUint32Exp{val: 4294967295, ok: true},
		"4294967296": atoUint32Exp{val: 0, ok: false},
		"4294967297": atoUint32Exp{val: 0, ok: false},
		"4294967298": atoUint32Exp{val: 0, ok: false},
		"4294967299": atoUint32Exp{val: 0, ok: false},
	}

	for str, exp := range expMap {
		val, ok := atoUint32([]byte(str))
		if val != exp.val || ok != exp.ok {
			t.Errorf("atoUint32(%s) want=%v got=%v", str, exp,
				atoUint32Exp{val: val, ok: ok})
		}
	}
}

func BenchmarkAtoUint32(b *testing.B) {
	bytes := []byte("4294967295")
	b.ReportAllocs()
	b.ResetTimer()
	for i := 0; i < b.N; i++ {
		_, ok := atoUint32(bytes)
		if !ok {
			b.Fatal("atoUint32() failed to parse")
			return
		}
	}
}

func BenchmarkStrconvAtoi(b *testing.B) {
	bytes := []byte("4294967295")
	b.ReportAllocs()
	b.ResetTimer()
	for i := 0; i < b.N; i++ {
		_, err := strconv.Atoi(string(bytes))
		if err != nil {
			b.Fatal("strconv.Atoi() failed to parse")
			return
		}
	}
}

var delimVec = map[string][]struct{
	fld int
	ok  bool
	exp string
}{
	"foo bar baz quux": {
		{0, true, "foo"},
		{1, true, "bar"},
		{2, true, "baz"},
		{3, true, "quux"},
		{4, false, ""},
		{5, false, ""},
	},
	"  foo \t \n  bar\r  \t   baz    quux\n\r\n\n": {
		{0, true, "foo"},
		{1, true, "bar"},
		{2, true, "baz"},
		{3, true, "quux"},
		{4, false, ""},
		{5, false, ""},
	},
	"foobarbazquux": {
		{0, true, "foobarbazquux"},
		{1, false, ""},
		{2, false, ""},
	},
	" \t\n\r foobarbazquux\r\n\n\n": {
		{0, true, "foobarbazquux"},
		{1, false, ""},
		{2, false, ""},
	},
	"foobarbazquux     ": {
		{0, true, "foobarbazquux"},
		{1, false, ""},
		{2, false, ""},
	},
	"     foobarbazquux": {
		{0, true, "foobarbazquux"},
		{1, false, ""},
		{2, false, ""},
	},
	"": {
		{0, false, ""},
		{1, false, ""},
		{2, false, ""},
	},
}
	
func TestFieldNDelims(t *testing.T) {
	s := []byte("foo bar baz quux")
	_, _, ok := fieldNDelims(s, 4)
	if ok {
		t.Error("fieldNDelims() did not fail for N out of range")
	}
	_, _, ok = fieldNDelims(s, -1)
	if ok {
		t.Error("fieldNDelims() did not fail for N < 0")
	}
	_, _, ok = fieldNDelims(nil, 0)
	if ok {
		t.Error("fieldNDelims() did not fail the nil slice")
	}

	for str, expVec := range delimVec {
		sl := []byte(str)
		for _, exp := range expVec {
			min, max, ok := fieldNDelims(sl, exp.fld)
			if ok != exp.ok {
				t.Errorf("fieldNdelims(%s, %d) ok want=%v "+
					"got=%v", str, exp.fld, exp.ok, ok)
			}
			if exp.ok {
				f := string(sl[min:max])
				if f != exp.exp {
					t.Errorf("fieldNDelims(%s, %d) want=%v"+
						" got=%v", str, exp.fld,
						exp.exp, f)
				}
			}
		}
	}
}

func BenchmarkFieldNDelims(b *testing.B) {
	bytes := []byte("foo bar baz quux")
	b.ReportAllocs()
	b.ResetTimer()
	for i := 0; i < b.N; i++ {
		_, _, ok := fieldNDelims(bytes, 3)
		if !ok {
			b.Fatal("fieldNDelims() failed to parse")
			return
		}
	}
}

func BenchmarkBytesFields(b *testing.B) {
	s := []byte("foo bar baz quux")
	b.ReportAllocs()
	b.ResetTimer()
	for i := 0; i < b.N; i++ {
		_ = bytes.Fields(s)
	}
}
